
/*
 *  replace.c
 *  Copyright (c) 1999 by Brad Proctor.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 */

/* 
 * $Id: replace.c,v 1.5 1999/06/19 20:16:04 brad Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "version.h"

#define BUF_SIZE  65536

#define ISALNUM   0xFF
#define ISALPHA   0xFE
#define ISCNTRL   0xFD
#define ISDIGIT   0xFC
#define ISGRAPH   0xFB
#define ISLOWER   0xFA
#define ISPRINT   0xF9
#define ISPUNCT   0xF8
#define ISSPACE   0xF7
#define ISUPPER   0xF6
#define ISXDIGIT  0xF5

#define invalid() \
{                 \
   fprintf(stderr, "%s: Invalid option: `%s'\n", argv[0], argv[i]);     \
   fprintf(stderr, "Try `%s --help' for more information.\n", argv[0]); \
   return (-1);   \
}

static char help_text[] =
{
   "Usage: replace [option] string1 string2\n\n"
   "  -i         Ignore case when comparing\n"
   "  --help     Print this help message\n"
   "  --version  Print the program version\n\n"
   "Report bugs to fredlie@sprynet.com"
};

static char flag_case = 0;     /* Set if case doesn't matter */
static char cbuf[BUF_SIZE];

static int string_compare(const char *s1, const char *s2)
{
   while(1)
   {
      switch((unsigned char)*s1)
      {
         case '\0':  
            return (0);
         case ISALNUM:
            if(!isalnum(*s2)) return (-1);
            break;
         case ISALPHA:
            if(!isalpha(*s2)) return (-1);
            break;
         case ISCNTRL:
            if(!iscntrl(*s2)) return (-1);
            break;
         case ISDIGIT:
            if(!isdigit(*s2)) return (-1);
            break;
         case ISGRAPH:
            if(!isgraph(*s2)) return (-1);
            break;
         case ISLOWER:
            if(!islower(*s2)) return (-1);
            break;
         case ISPRINT:
            if(!isprint(*s2)) return (-1);
            break;
         case ISPUNCT:
            if(!ispunct(*s2)) return (-1);
            break;
         case ISSPACE:
            if(!isspace(*s2)) return (-1);
            break;
         case ISUPPER:
            if(!isupper(*s2)) return (-1);
            break;
         case ISXDIGIT:
            if(!isxdigit(*s2)) return (-1);
            break;
         default:
            if(flag_case)
            {
               if(tolower(*s1) != tolower(*s2))
                  return (-1);
            }
            else
            {
               if(*s1 != *s2)
                  return (-1);
            }
      }
      ++s1, ++s2;
   }
}

/*
 * Convert characters with '\' in them to 
 * what they are suppost to be.
 */

static char *fix_string(char *ptr)
{
   int i;
   char *ptr1, *buf;

   if((buf = (char *)malloc(strlen(ptr))) == NULL)
   {
      printf("Out of memory\n");
      exit(-1);
   }
   ptr1 = buf;

   while(*ptr)
   {
      if(*ptr == '\\')
      {
         switch(*++ptr)
         {
            case 'a':   *ptr1++ = '\a';   break;
            case 'b':   *ptr1++ = '\b';   break;
            case 'f':   *ptr1++ = '\f';   break;
            case 'n':   *ptr1++ = '\n';   break;
            case 'r':   *ptr1++ = '\r';   break;
            case 't':   *ptr1++ = '\t';   break;
            case 'v':   *ptr1++ = '\v';   break;
            case '\'':  *ptr1++ = '\'';   break;
            case '\"':  *ptr1++ = '\"';   break;
            case '\\':  *ptr1++ = '\\';   break;
            case '\?':  *ptr1++ = '\?';   break;
            case 'x':
               i = 0;
               while(1)
               {
                  ++ptr;
                  if(isdigit(*ptr))
                     i = (i << 4) + *ptr - '0';
                  else if(*ptr >= 'a' && *ptr <= 'f')
                     i = (i << 4) + *ptr - 'a' + 10;
                  else if(*ptr >= 'A' && *ptr <= 'F')
                     i = (i << 4) + *ptr - 'A' + 10;
                  else
                     break;   
               }
               *ptr1++ = i;
               break;
            case '0':
               i = 0;
               while(1)
               {
                  ++ptr;
                  if(*ptr >= '0' && *ptr <= '7')
                     i = (i << 3) + *ptr - '0';
                  else
                     break;
               }
               *ptr1++ = i;
               break;
            case 'i':
               ++ptr;
               if(*ptr != 's')
               {
                  *ptr1++ = '\\';
                  *ptr1++ = 'i';
                  *ptr1++ = *ptr;
               }
               ++ptr;
               if(!strncmp(ptr, "alnum", 5))
                  *ptr1++ = ISALNUM, ptr += 4;
               else if(!strncmp(ptr, "alpha", 5))
                 *ptr1++ = ISALPHA, ptr += 4;
               else if(!strncmp(ptr, "cntrl", 5))
                  *ptr1++ = ISCNTRL, ptr += 4;
               else if(!strncmp(ptr, "digit", 5))
                  *ptr1++ = ISDIGIT, ptr += 4;
               else if(!strncmp(ptr, "graph", 5))
                  *ptr1++ = ISGRAPH, ptr += 4;
               else if(!strncmp(ptr, "lower", 5))
                  *ptr1++ = ISLOWER, ptr += 4;
               else if(!strncmp(ptr, "print", 5))
                  *ptr1++ = ISPRINT, ptr += 4;
               else if(!strncmp(ptr, "punct", 5))
                  *ptr1++ = ISPUNCT, ptr += 4;
               else if(!strncmp(ptr, "space", 5))
                  *ptr1++ = ISSPACE, ptr += 4;
               else if(!strncmp(ptr, "upper", 5))
                  *ptr1++ = ISUPPER, ptr += 4;
               else if(!strncmp(ptr, "xdigit", 6))
                  *ptr1++ = ISXDIGIT, ptr += 5;
               else
               {
                  *ptr1++ = '\\';
                  *ptr1++ = 'i';
                  *ptr1++ = *ptr;
               }
               break;
            default:
               *ptr1++ = '\\';
               *ptr1++ = *ptr;
         }
         ++ptr;
      }
      else
         *ptr1++ = *ptr++;
   }
   *ptr1 = '\0';
   return (buf);
}

/*
 * Main
 */

int main(int argc, char *argv[])
{
	char *pat1 = 0, *pat2 = 0;
   char *p, *ptr, buf[100];
	int c, i;
	   
	for(i = 1; i < argc; ++i)
	{
	   ptr = argv[i];
	   if(*ptr == '-')
	   {
         switch(*++ptr)
         {
            case '-':
               ++ptr;
               if(!strcmp(ptr, "help"))
               {
                  puts(help_text);
                  return (0);
               }
               else if(!strcmp(ptr, "version"))
               {
                  puts("replace (butils) "VERSION_STRING);
                  return (0);
               }
               else
                  invalid();               
               break;
            case 'i':
               if(*++ptr != '\0')
                  invalid();
               flag_case = 1;
               break;
            default:
               invalid();
         }
	   }
	   else if(!pat1)
	      pat1 = fix_string(argv[i]);
	   else if(!pat2)
	      pat2 = fix_string(argv[i]);
      else
         invalid();
	}
   if(!pat1 || !pat2)
   {
      fprintf(stderr, "%s: Two strings are required.\n", argv[0]);
      puts("Try 'replace --help' for more information.");
      return (-1);
   }
   
	i = strlen(pat1);
   while((c = fread(cbuf, 1, BUF_SIZE, stdin)) > 0)
   {
      p = cbuf;
      while(p - cbuf < c)
      {
   	   if((flag_case && tolower(*p) == tolower(*pat1))
   	      || (!flag_case && *p == *pat1)
   	      || (unsigned char)*pat1 >= (unsigned char)ISXDIGIT)
         {
            ptr = buf;
            *ptr = *p;
            while(ptr - buf < i)
            {
               if(*++p == '\0')
               {
                  if((c = fread(cbuf, 1, BUF_SIZE, stdin)) <= 0)
                     break;
                  p = cbuf;
               }
               *++ptr = *p;
            }
            *ptr = '\0';
            if(!string_compare(pat1, buf))
               fputs(pat2, stdout);
            else
               fputs(buf, stdout);
         }
         else
            putchar(*p++);
      }            
   }
	return (0);
}
